/* This program "compiles" the hook configuration file.         */

/* Copyright (C) 2002, David W Noon.  All rights reserved.      */

/* You may use this code freely. You may also distribute it     */
/* freely, provided the original copyright caveat is preserved  */
/* and no charge is levied beyond the price of its distribution */
/* medium.                                                      */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define INCL_WININPUT
#define INCL_WINERRORS
#define INCL_DOSNLS
#define INCL_DOSERRORS
#include <os2.h>

/* We are also using the 32-bit CRC algorithm from CCITT X.25. */
#include "crc32.h"

/* We want an array of scancodes, too. */
#include "scancode.h"

/* Define key translation data structures. */
#include "keytrans.h"

/* This function searches the scan_code[] array for a given scancode. */
static PCSCANCODE FindScancode(UCHAR scancode)
{
   int l = 0, r = sizeof(scan_code)/sizeof(scan_code[0]) - 1, m;

   /* Since the scancodes are sorted, we can use a binary search. */
   if (scan_code[0].scancode == scancode)
      m = 0;
   else if (scan_code[r].scancode == scancode)
      m = r;
   else
   {
      while (r >= l)
      {
         m = (l + r) >> 1;
         if (scan_code[m].scancode == scancode)
            break;

         if (scan_code[m].scancode < scancode)
            l = m + 1;
         else
            r = m - 1;
      }
   }

   return r < l ? NULL : &scan_code[m];
}

/* Subroutine to convert a mnemonic string to a CHRMSG structure. */
static void KeystrokeMnemonicToDownstroke(PCSZ key, PCHRMSG chrmsg)
{
   PCSZ hyphen, p;
   int i, l, r, m, shift_sub;
   const int max_clever_key = sizeof(clever_key)/sizeof(clever_key[0]),
         max_scan_code = sizeof(scan_code)/sizeof(scan_code[0]);

   /* Nullify the character structure. */
   memset(chrmsg, 0x00, sizeof(CHRMSG));

   /* If there is no point ... */
   if (key == NULL)
      return;

   /* Process the mnemonic string. Length is the first clue. */
   switch (strlen(key))
   {
   case 0:
      /* Not worth the trouble. */
      break;

   case 1:
      chrmsg->fs = KC_CHAR;
      chrmsg->vkey = 0U;
      chrmsg->cRepeat = 1U;
      chrmsg->chr = key[0];
      for (i = 0; i < max_scan_code; ++i)
      {
         if (scan_code[i].chr_vkey[0] == 1U &&
               scan_code[i].chr_code[0] == chrmsg->chr)
         {
            chrmsg->scancode = scan_code[i].scancode;
            chrmsg->fs |= KC_SCANCODE;
            break;
         }
      }
      break;

   default:
      hyphen = strchr(key, '-');
      if (hyphen != NULL && hyphen - key <= 3)
      {
         /* Parse the Ctl, Alt and Shift prefixes supplied. */
         for (p = key; p < hyphen; ++p)
         {
            switch (*p)
            {
            case 'C':
               chrmsg->fs |= KC_CTRL;
               break;

            case 'A':
               chrmsg->fs |= KC_ALT;
               break;

            case 'S':
               chrmsg->fs |= KC_SHIFT;
               break;

            default:
               /* Duff input! */
               chrmsg->fs = 0U;
               return;
               break;
            }
         }

         /* Now process the text after the hyphen. */
         p = hyphen + 1;
         switch (strlen(p))
         {
         case 0:
            /* All that good work, down the drain. */
            chrmsg->fs = 0U;
            break;

         case 1:
            /* Those who use decision tables will understand this next run of code. */
            if (chrmsg->fs & KC_ALT)
               if (chrmsg->fs & KC_CTRL)
                  if (chrmsg->fs & KC_SHIFT)
                     shift_sub = 6;
                  else
                     shift_sub = 3; /* Alt+Ctrl gives the chr value same as Alt */
               else if (chrmsg->fs & KC_SHIFT)
                  shift_sub = 5;
               else
                  shift_sub = 3;
            else if (chrmsg->fs & KC_CTRL)
               if (chrmsg->fs & KC_SHIFT)
                  shift_sub = 4;
               else
                  shift_sub = 2;
            else if (chrmsg->fs & KC_SHIFT)
               shift_sub = 1;
            else
               shift_sub = 0;

            /* Since the scancodes are not sorted on chr_code[0] we have to use a dumb search. */
            for (i = 0; i < max_scan_code; ++i)
            {
               if (scan_code[i].chr_vkey[shift_sub] == 1U && scan_code[i].chr_code[0] == *p)
               {
                  chrmsg->scancode = scan_code[i].scancode;
                  chrmsg->chr = *p;
                  chrmsg->fs |= KC_SCANCODE;
                  break;
               }
            }

            chrmsg->cRepeat = 1U;
            break;

         default:
            /* Set the virtual key and scancode. */
            for (i = 0; i < max_clever_key; ++i)
            {
               if (strcmp(clever_key[i].key_mnemonic, p) == 0)
               {
                  chrmsg->fs |= KC_SCANCODE;
                  chrmsg->cRepeat = 1U;
                  chrmsg->scancode = clever_key[i].scancode;
                  chrmsg->vkey = clever_key[i].vkey_code;
                  if (chrmsg->scancode < 0x20) /* This could be wrong. */
                  {
                     /* This is for Tab, Backspace, etc. */
                     chrmsg->chr = chrmsg->scancode;
                  }
                  else
                  {
                     /* A proper virtual key. */
                     chrmsg->fs |= KC_VIRTUALKEY;

                     if ((chrmsg->fs & (KC_ALT | KC_CTRL | KC_SHIFT)) == KC_SHIFT)
                     {
                        const PCSCANCODE scan_code_entry = FindScancode(chrmsg->scancode);

                        /* Extract the shift scancode. */
                        chrmsg->chr = scan_code_entry->chr_code[1];
                        chrmsg->fs |= KC_CHAR;
                     }
                     else
                        chrmsg->chr = (USHORT)((chrmsg->fs & KC_ALT) ? clever_key[i].alt_scancode :
                              (chrmsg->fs & KC_CTRL) ? clever_key[i].ctrl_scancode :
                              clever_key[i].scancode) << 8;
                  }
                  break;
               }
            }
            break;
         }
      }
      else if (key[0] == 'F' && strlen(key) <= 3 && strcspn(&key[1], "0123456789") == 0)
      {
         const ULONG F_key = strtoul(&key[1], NULL, 10);

         if (F_key >= 1UL && F_key <= 24UL) /* It's a valid function key. */
         {
            chrmsg->fs = KC_VIRTUALKEY | KC_SCANCODE;
            chrmsg->cRepeat = 1U;
            chrmsg->scancode = clever_key[F_key-1].scancode;
            chrmsg->vkey = clever_key[F_key-1].vkey_code;
         }
      }
      break;
   }

   return;
}

/* This subroutine converts a downstroke CHRMSG to its corresponding upstroke. */
static void DownstrokeToUpstroke(PCHRMSG down, PCHRMSG up)
{
   /* Simple copy to get things going. */
   memcpy(up, down, sizeof(CHRMSG));

   /* Add the key up flag. */
   up->fs |= KC_KEYUP;

   /* Fiddle the character code, irrationally, based on the scan code. */
   if ((up->fs & (KC_SCANCODE | KC_VIRTUALKEY)) == KC_SCANCODE)
   {
      const PCSCANCODE scan_code_entry = FindScancode(up->scancode);
      int shift_sub;

      if (up->fs & KC_ALT)
         if (up->fs & KC_CTRL)
            if (up->fs & KC_SHIFT)
               shift_sub = 6;
            else
               shift_sub = 3; /* Alt+Ctrl gives the chr value same as Alt */
         else if (up->fs & KC_SHIFT)
            shift_sub = 5;
         else
            shift_sub = 3;
      else if (up->fs & KC_CTRL)
         if (up->fs & KC_SHIFT)
            shift_sub = 4;
         else
            shift_sub = 2;
      else if (up->fs & KC_SHIFT)
         shift_sub = 1;
      else
         shift_sub = 0;

      up->chr = ((USHORT) up->scancode) << 8 | scan_code_entry->chr_code[shift_sub];
   }
   else /* It is a virtual key */
   {
      if (HIBYTE(up->chr) == 0) /* Put the base scancode into the high byte. */
      {
         up->chr |= ((USHORT) up->scancode) << 8;
         up->fs &= ~ KC_CHAR;
      }
   }

   return;
}

/* Comparison routine for qsort() of Programs array. */
static int CRC_compare(const void * a, const void * b)
{
   register const ULONG CRC_a = ((PPGMTRANS) a)->CRC_value;
   register const ULONG CRC_b = ((PPGMTRANS) b)->CRC_value;

   return CRC_a > CRC_b ? 1 : CRC_a < CRC_b ? -1 : 0;
}

int main(int argc, char ** argv)
{
   ULONG No_translations = 0UL, No_programs = 0UL, trans_no;
   PKEYTRANS Translations = NULL;
   PPGMTRANS Programs = NULL;
   size_t len;
   int    n;
   PUCHAR p, q;
   FILE * const input_file = fopen(argc < 2 ? "inpthook.txt" : argv[1], "r");
   UCHAR  line_buffer[512], executable_name[CCHMAXPATH];
   static const UCHAR delimiters[] = " \x09";
#if !CASE_SENSITIVE_PROGRAM_NAMES
   COUNTRYCODE country_code;
#endif

   if (input_file != NULL)
   {
      memset(line_buffer, 0x00, sizeof line_buffer);
      memset(executable_name, 0x00, sizeof executable_name);

      /* Read the first line from the configuration file. */
      if (fgets(line_buffer, sizeof line_buffer, input_file) != NULL)
      {
         /* The first line contains the number of key translation pairs. */
         No_translations = strtoul(line_buffer, NULL, 0);
         /* We need both a downstroke and an upstroke, so double the allocation. */
         No_translations <<= 1;

         /* Allocate the array of key translation pairs. */
         Translations = (PKEYTRANS) calloc(No_translations, sizeof(KEYTRANS));
         if (Translations != NULL)
         {
            /* Loop through file putting pairs into the newly allocated array. */
            for (trans_no = 0UL; trans_no < No_translations &&
                  fgets(line_buffer, sizeof line_buffer, input_file) != NULL; trans_no += 2UL)
            {
               /* Trim off trailing \n character. */
               line_buffer[strlen(line_buffer)-1] = '\0';

               /* Extract the "input" mnemonic and convert it to CHRMSG structures. */
               KeystrokeMnemonicToDownstroke(strtok(line_buffer, delimiters),
                     &Translations[trans_no].key_in);
               DownstrokeToUpstroke(&Translations[trans_no].key_in,
                     &Translations[trans_no+1].key_in);

               /* Extract the "output" mnemonic and convert it to CHRMSG structures. */
               KeystrokeMnemonicToDownstroke(strtok(NULL, delimiters),
                     &Translations[trans_no].key_out);
               DownstrokeToUpstroke(&Translations[trans_no].key_out,
                     &Translations[trans_no+1].key_out);
            }
         }

         /* Read the next line from the configuration file. */
         if (fgets(line_buffer, sizeof line_buffer, input_file) != NULL)
         {
            /* The first line contains the number of program definitions. */
            No_programs = strtoul(line_buffer, NULL, 0);

            /* Allocate the array of program specifications. */
            Programs = (PPGMTRANS) calloc(No_programs, sizeof(PGMTRANS));
            if (Programs != NULL)
            {
               /* Loop through file putting programs into the newly allocated array. */
               for (trans_no = 0UL; trans_no < No_programs &&
                     fgets(line_buffer, sizeof line_buffer, input_file) != NULL; ++trans_no)
               {
                  /* Trim off trailing \n character. */
                  line_buffer[strlen(line_buffer)-1] = '\0';

                  /* Extract the program's path/filename. */
                  switch (line_buffer[0])
                  {
                  case '"':
                  case '\'':
                     /* The path/filename is quoted. */

                     /* Scan for the trailing delimiter. */
                     p = strchr(&line_buffer[1], line_buffer[0]);

                     /* p should never be NULL. */
                     if (p != NULL && (p-&line_buffer[1]) < CCHMAXPATH)
                     {
                        len = p - line_buffer - 1;
                        memcpy(Programs[trans_no].Executable_name, &line_buffer[1], len);
                     }
                     else
                        continue;
                     break;

                  default:
                     /* The path/filename ends with a blank or tab. */
                     len = strcspn(line_buffer, " \x09");
                     if (len <= CCHMAXPATH)
                     {
                        memcpy(Programs[trans_no].Executable_name, line_buffer, len);
                        p = &line_buffer[len];
                     }
                     else
                        continue;
                     break;
                  }

#if !FULL_PATH_PROGRAM_NAMES
                  q = strrchr(Programs[trans_no].Executable_name, '\\');
                  if (q != NULL)
                     strcpy(Programs[trans_no].Executable_name, q+1);
#endif

                  len = strlen(Programs[trans_no].Executable_name);

#if !CASE_SENSITIVE_PROGRAM_NAMES
                  /* Make the executable name capitals. */
                  country_code.country = 0UL;
                  country_code.codepage = 0UL;
                  DosMapCase(len, &country_code, Programs[trans_no].Executable_name);
#endif

                  /* Take the CRC32 of the executable's path/filename. */
                  Programs[trans_no].CRC_value = CRC32(Programs[trans_no].Executable_name, len);

                  /* Scan over subsequent spaces and tabs. */
                  len = strspn(p+1, delimiters);
                  p += len + 1;

                  /* Now scan for commas separating translation numbers. */
                  for (n = 0; n < 15 && (q = strchr(p, ',')) != NULL; ++n)
                  {
                     *q = '\0';
                     Programs[trans_no].Translations[n] = (strtoul(p, NULL, 0) - 1UL) << 1;
                     p = q + 1;
                  }
                  Programs[trans_no].Translations[n] = (strtoul(p, NULL, 0) - 1UL) << 1;
                  Programs[trans_no].No_translations = n + 1;
               }

               /* Now sort the array by ascending CRC32 value. */
               if (No_programs > 1)
                  qsort(Programs, No_programs, sizeof(PGMTRANS), &CRC_compare);
            }
         }
      }

      fclose(input_file);

      /* Did we get enough to build a sensible configuration? */
      if (No_translations > 0UL && No_programs > 0UL)
      {
         FILE * const config = fopen("inpthook.cfg", "wb");

         fwrite(&No_translations, sizeof(ULONG), 1, config);
         fwrite(&No_programs, sizeof(ULONG), 1, config);
         fwrite(Translations, sizeof(KEYTRANS), No_translations, config);
         fwrite(Programs, sizeof(PGMTRANS), No_programs, config);
         fclose(config);

         fputs("Configuration file successfully written.", stdout);
      }
      else
         fputs("Failed to build configuration.", stderr);

      /* Release the memory. */
      free(Translations);
      free(Programs);
   }
   else
      fputs("Unable to open input file.", stderr);

   return 0;
}
