/* The routine in this DLL is a filter for WM_CHAR messages.    */

/* Copyright (C) 2002, David W Noon.  All rights reserved.      */

/* You may use this code freely. You may also distribute it     */
/* freely, provided the original copyright caveat is preserved  */
/* and no charge is levied beyond the price of its distribution */
/* medium.                                                      */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <process.h>

#define INCL_WINMESSAGEMGR
#define INCL_WINWINDOWMGR
#define INCL_WININPUT
#define INCL_WINHOOKS
#define INCL_WINERRORS
#define INCL_DOSPROFILE
#define INCL_DOSNLS
#define INCL_DOSERRORS
#include <os2.h>

/* We are also using the 32-bit CRC algorithm from CCITT X.25. */
#include "crc32.h"

/* We want an array of scancodes, too. */
#include "scancode.h"

/* Define key translation data structures. */
#include "keytrans.h"

/* Size of the stack for our set-up thread. */
static const unsigned long stack_size = 32UL << 10;

ULONG No_translations = 0UL, No_programs = 0UL;
PKEYTRANS Translations = NULL;
PPGMTRANS Programs = NULL;

/* The initialization routine will be a separate thread */
static void DLL_setup(void * thread_parm)
{
   FILE * config;

   config = fopen(thread_parm == NULL ? "INPTHOOK.CFG" : (char *) thread_parm, "rb");
   if (config == NULL && thread_parm == NULL)
   {
      char file_name[CCHMAXPATH];

      /* See if our file is somewhere on the DPATH directory list. */
      if (DosSearchPath(SEARCH_IGNORENETERRS | SEARCH_ENVIRONMENT, "DPATH",
            "INPTHOOK.CFG", file_name, sizeof file_name) == NO_ERROR)
         config = fopen(file_name, "rb");
   }

   if (config != NULL)
   {
      char config_buffer[16384];

      setvbuf(config, config_buffer, _IOFBF, sizeof config_buffer);

      fread(&No_translations, sizeof(ULONG), 1, config);
      fread(&No_programs, sizeof(ULONG), 1, config);

      Translations = (PKEYTRANS) calloc(No_translations, sizeof(KEYTRANS));
      Programs = (PPGMTRANS) calloc(No_programs, sizeof(PGMTRANS));

      if (Translations == NULL || Programs == NULL)
      {
         /* Forget about it. */
         free(Translations);
         free(Programs);
         No_translations = 0UL;
         No_programs = 0UL;
      }
      else
      {
         fread(Translations, sizeof(KEYTRANS), No_translations, config);
         fread(Programs, sizeof(PGMTRANS), No_programs, config);
      }

      fclose(config);
   }

   return;
}

/* Our custom clean-up routine *before* the CRTL shuts down. */
static void DLL_cleanup(void)
{
   /* Return any memory allocations to a potentially shared heap */
   free(Translations);
   free(Programs);

   return;
}

#if defined(__IBMC__) || defined(__BORLANDC__) || defined(__EMX__)
/* This breaks the naming rules of C, but it is Watcom-compatible. */
HMODULE __hmodule = NULLHANDLE;

/* Prototypes for the CRTL start-up/shut-down routines. */
int _CRT_init(void);
#if !defined(__IBMC__) || defined(STATIC_LINK)
void _CRT_term(void);
#endif
void __ctordtorInit(void);
void __ctordtorTerm(void);

/* Typical initialize/terminate routine for a DLL. */
unsigned long _System _DLL_InitTerm(unsigned long mod_handle, unsigned long term_flag)
{
   if (term_flag)
   {
      /* Call our own clean-up routine first. */
      DLL_cleanup();

      __ctordtorTerm();
#if !defined(__IBMC__) || defined(STATIC_LINK)
      _CRT_term();
#endif
   }
   else
   {
      if (_CRT_init() != 0)
         return 0;
      __ctordtorInit();

      /* Make our module handle available, just like Watcom does. */
      __hmodule = (HMODULE) mod_handle;

      /* Call our set-up routine on a separate thread. */
#ifdef __BORLANDC__
      /* Borland uses a non-standard argument list for this function! */
      _beginthread(&DLL_setup, stack_size, NULL);
#else
      _beginthread(&DLL_setup, NULL, stack_size, NULL);
#endif
   }

   return 1UL;
}
#elif defined(__WATCOMC__)
#if __WATCOMC__ >= 1100
/* For V11.0 onwards, we use LibMain. */
unsigned long LibMain(unsigned long mod_handle, unsigned long term_flag)
{
   if (term_flag)
   {
      /* Call our own clean-up routine. */
      DLL_cleanup();
   }
   else
   {
      /* Call our set-up routine on a separate thread. */
      _beginthread(&DLL_setup, NULL, stack_size, NULL);
   }
   return 1UL;
}
#else
/* For V10.6 and earlier, use Watcom specials. */
int __dll_initialize(void)
{
   /* Call our set-up routine on a separate thread. */
   _beginthread(&DLL_setup, NULL, stack_size, NULL);

   /* Set a pointer to our clean-up routine, so it runs before CRTL termiantion. */
   atexit(DLL_cleanup);

   return 1;
}

int __dll_terminate(void)
{
   return 1;
}
#endif
#else
#error "Unable to determine C compiler type."
#endif

/* Function to determine the path/filename of the main program that owns a window. */
static APIRET PgmNameFromHWND(HWND hwnd, ULONG exec_name_size, PSZ exec_name)
{
   ULONG buffer[256]; /* Is 1 KiB enough? */
   PID  pid;
   TID  tid;
   APIRET rc;

   if (WinQueryWindowProcess(hwnd, &pid, &tid))
   {
      /* How fast/slow is this API call? */
      rc = DosQuerySysState(QS_PROCESS, 0UL, pid, 0UL, buffer, sizeof buffer);
      if (rc == NO_ERROR)
      {
         /* The data mapping of these performance profiling structures is butt-ugly. */
         rc = DosQueryModuleName((HMODULE)((qsPrec_t *)(buffer[0]+sizeof(qsGrec_t)))->hMte,
               exec_name_size, exec_name);
      }
   }
   else
      rc = PMERR_INVALID_HWND;

   return rc;
}

/* This function determines the name of the program that created the target window */
/* and then sees if that progam is one whose input we should filter.               */
static PPGMTRANS FindProgram(HWND hwnd)
{
   ULONG CRC_value;
   int l, r, m;
   size_t len;
   UCHAR executable_name[CCHMAXPATH];
#if !CASE_SENSITIVE_PROGRAM_NAMES
   COUNTRYCODE country_code;
#endif

   /* Find out who the message recipient is. */
   if (PgmNameFromHWND(hwnd, sizeof executable_name, executable_name) != NO_ERROR)
      return NULL;

#if !FULL_PATH_PROGRAM_NAMES
   strcpy(executable_name, strrchr(executable_name, '\\')+1);
#endif

   len = strlen(executable_name);

#if !CASE_SENSITIVE_PROGRAM_NAMES
   /* Make the executable name capitals. */
   country_code.country = 0UL;
   country_code.codepage = 0UL;
   DosMapCase(len, &country_code, executable_name);
#endif

   /* Do the initial search by CRC, as it's faster than matching strings. */
   CRC_value = CRC32(executable_name, len);

   /* Search for a matching CRC using binary search. */
   r = No_programs - 1U; /* Right extremum */
   /* Check the extrema first. */
   if (Programs[0].CRC_value == CRC_value)
      m = 0;
   else if (Programs[r].CRC_value == CRC_value)
      m = r;
   else
   {
      /* Left extremum is 0, of course. */
      l = 0;

      /* Now the binary search proper. */
      while (r >= l)
      {
         /* Take a midpoint value. */
         m = (l + r) >> 1;
         /* See if it matches. */
         if (Programs[m].CRC_value == CRC_value)
            break;

         /* Adjust either the left or right extremum. */
         if (Programs[m].CRC_value < CRC_value)
            l = m + 1;
         else
            r = m - 1;
      }

      /* No CRC match, we're done. */
      if (r < l)
         return NULL;
   }

   /* Now perform an exact string match on only those with CRC matches. */
   for (l = m; l >= 0 && Programs[l].CRC_value == CRC_value; --l)
      if (strcmp(Programs[l].Executable_name, executable_name) == 0)
         return &Programs[l];
   for (r = m + 1; r < No_programs && Programs[r].CRC_value == CRC_value; ++r)
      if (strcmp(Programs[r].Executable_name, executable_name) == 0)
         return &Programs[r];

   /* No match. */
   return NULL;
}

/* This bit mask maps the significant bits in the flags field of a CHRMSG. */
static const USHORT bit_mask = KC_KEYUP | KC_ALT | KC_CTRL | KC_SHIFT | KC_SCANCODE | KC_VIRTUALKEY | KC_CHAR;

/* This function compares 2 CHRMSG structures for logical equality. */
static BOOL MatchCHRMSG(PCHRMSG a, PCHRMSG b)
{
   return (a->fs & bit_mask) == (b->fs & bit_mask) &&
         (!(a->fs & KC_SCANCODE) || a->scancode == b->scancode) &&
         (!(a->fs & KC_VIRTUALKEY) || a->vkey == b->vkey) &&
         (!(a->fs & KC_CHAR) || a->chr == b->chr ||
            (a->fs & KC_KEYUP) && HIBYTE(a->chr) == HIBYTE(b->chr));
}

BOOL EXPENTRY KeystrokeFilterHook(HAB anchor_block, PQMSG pqmsg, ULONG removal)
{
   PPGMTRANS pgm_entry;
   int i;
   CHRMSG msg_chr;

   /* Only process WM_CHAR messages. */
   if (pqmsg->msg != WM_CHAR ||
         No_translations == 0U ||
         No_programs == 0U ||
         removal == PM_NOREMOVE)
      return FALSE;

   /* Only process for known programs. */
   if ((pgm_entry = FindProgram(pqmsg->hwnd)) == NULL)
      return FALSE;

   /* Extract the message's CHRMSG values. */
   msg_chr = *CHARMSG(&(pqmsg->msg));

   /* Check the list for matching input keystrokes. */
   for (i = 0; i < pgm_entry->No_translations; ++i)
   {
      PKEYTRANS trans = &Translations[pgm_entry->Translations[i]];

      /* See how the downstroke goes. */
      if (MatchCHRMSG(&(trans->key_in), &msg_chr))
      {
         msg_chr.fs &= ~ bit_mask;
         msg_chr.fs |= trans->key_out.fs & bit_mask;

         msg_chr.scancode = trans->key_out.scancode;
         msg_chr.chr = trans->key_out.chr;
         msg_chr.vkey = trans->key_out.vkey;

         /* Put the revised message back in the parameter area. */
         *CHARMSG(&(pqmsg->msg)) = msg_chr;
      
         break;
      }

      /* No match. Try the upstroke, then. */
      ++trans;
      if (MatchCHRMSG(&(trans->key_in), &msg_chr))
      {
         msg_chr.fs &= ~ bit_mask;
         msg_chr.fs |= trans->key_out.fs & bit_mask;

         msg_chr.scancode = trans->key_out.scancode;
         msg_chr.chr = trans->key_out.chr;
         msg_chr.vkey = trans->key_out.vkey;


         /* Put the revised message back in the parameter area. */
         *CHARMSG(&(pqmsg->msg)) = msg_chr;
      
         break;
      }
   }

   return FALSE;
}
