/* Program to act as test harness for PM hook routines.         */

/* Copyright (C) 2002, David W Noon.  All rights reserved.      */

/* You may use this code freely. You may also distribute it     */
/* freely, provided the original copyright caveat is preserved  */
/* and no charge is levied beyond the price of its distribution */
/* medium.                                                      */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <process.h>

#define INCL_WINHOOKS
#define INCL_WININPUT
#define INCL_WINMESSAGEMGR
#define INCL_WINWINDOWMGR
#define INCL_WINERRORS
#define INCL_DOSMODULEMGR
#define INCL_DOSPROCESS
#define INCL_DOSSEMAPHORES
#define INCL_DOSERRORS
#include <os2.h>

#include "testhook.h"

/* The threads can communicate using this structure. */
typedef struct
{
   volatile BOOL32 terminate_thread;
   HEV   event_sem;
   UCHAR window_text[2000];
} THREADAREA, * PTHREADAREA;

/* A background thread subroutine to log the keystrokes to a file. */
static void FileThread(void * thread_parm)
{
   PTHREADAREA const thread_area = (PTHREADAREA) thread_parm;
   FILE * const log_file = fopen("testhook.log", "w");
   ULONG post_count;
   APIRET rc;
   UCHAR file_buffer[32768];

   if (log_file != NULL)
   {
      /* Use efficient buffering. */
      setvbuf(log_file, file_buffer, _IOFBF, sizeof file_buffer);

      fputs("Start of hook testing log.\n", log_file);

      rc = DosWaitEventSem(thread_area->event_sem, SEM_INDEFINITE_WAIT);
      while (!thread_area->terminate_thread && rc == NO_ERROR)
      {
         post_count = 0UL;
         if (DosResetEventSem(thread_area->event_sem, &post_count) != NO_ERROR)
            break;
         if (post_count > 1UL)
            fprintf(log_file, "%ul keystrokes were missed from log.\n", post_count-1UL);

         fputs(thread_area->window_text, log_file);
         fputc('\n', log_file);
         rc = DosWaitEventSem(thread_area->event_sem, SEM_INDEFINITE_WAIT);
      }
      /* Clear off any residual posts. */
      DosResetEventSem(thread_area->event_sem, &post_count);

      fputs("End of hook testing log.", log_file);
      fclose(log_file);
   }

   /* Tell them we've gone. */
   thread_area->terminate_thread = TRUE;

   return;
}

static MRESULT _System MainWndProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   PTHREADAREA const thread_area = (PTHREADAREA) WinQueryWindowPtr(hwnd, 0UL);
   HPS pres_space;
   RECTL client_area;
   CHRMSG chr;

   switch (msg)
   {
   case WM_CHAR:
      /* Extract the character code from the message parameters. */
      chr = *CHARMSG(&msg);

      sprintf(thread_area->window_text,
            "fs = 0x%.04X cRepeat = %u scancode = %u (0x%.02X) chr = 0x%.04X vkey = 0x%.04X",
            chr.fs, (USHORT) chr.cRepeat, (USHORT) chr.scancode, (USHORT) chr.scancode,
            chr.chr, chr.vkey);

      /* Let the background thread log the keystroke. */
      if (!thread_area->terminate_thread)
         DosPostEventSem(thread_area->event_sem);

      /* Force a full window paint. */
      WinInvalidateRect(hwnd, NULL, FALSE);

      return MRFROMLONG(0L);
      break;

   case WM_PAINT:
      /* Paint the client area. */
      pres_space = WinBeginPaint(hwnd, NULLHANDLE, &client_area);
      WinFillRect(pres_space, &client_area, CLR_BACKGROUND);
      if (strlen(thread_area->window_text) > 0)
         WinDrawText(pres_space, strlen(thread_area->window_text), thread_area->window_text,
               &client_area, 0L, 0L, DT_TEXTATTRS | DT_TOP | DT_LEFT);
      WinEndPaint(pres_space);

      return MRFROMLONG(0L);
      break;

   case WM_COMMAND:
      if (SHORT1FROMMP(mp1) == IDM_EXIT)
      {
         WinPostMsg(hwnd, WM_CLOSE, MPVOID, MPVOID);
         return MRFROMLONG(0L);
      }
      break;

   case WM_CLOSE:
      /* Tell the background thread we are closing down. */
      if (!thread_area->terminate_thread)
      {
         thread_area->terminate_thread = TRUE;
         DosPostEventSem(thread_area->event_sem);
      }
      break;

   default:
      break;
   }

   return WinDefWindowProc(hwnd, msg, mp1, mp2);
}

int main(int argc, char ** argv)
{
   HAB     anchor_block;
   HMQ     msg_queue;
   HWND    main_frame_window, main_client_window;
   HMODULE hook_dll;
   PFN     hook_routine;
   APIRET  rc;
   TID     file_thread;
   ULONG   Frame_flags = FCF_TITLEBAR | FCF_SIZEBORDER | FCF_MINMAX | FCF_SYSMENU |
         FCF_SHELLPOSITION | FCF_TASKLIST | FCF_ICON | FCF_ACCELTABLE | FCF_MENU;
   UCHAR   window_class[] = "Hook_test", window_title[] = "Test input hook", crap[CCHMAXPATH];
   QMSG    msg;
   THREADAREA thread_area;

   /* Basic PM housekeeping. */
   anchor_block = WinInitialize(0UL);
   msg_queue = WinCreateMsgQueue(anchor_block, 0UL);

   /* See if we should load the DLL. */
   if (argc > 1 && (argv[1][0] == '-' || argv[1][0] == '/') && stricmp(&argv[1][1], "noload") == 0)
   {
      rc = NO_ERROR;
      hook_dll = NULLHANDLE;
      hook_routine = NULL;
      memset(crap, 0x00, sizeof crap);
   }
   else
   {
      /* Load the DLL containing the hook routine. */
      rc = DosLoadModule(crap, sizeof crap, ".\\INPTHOOK.DLL", &hook_dll);
   }

   if (rc == NO_ERROR)
   {
      if (hook_dll != NULLHANDLE)
      {
         /* Our input hook is always at ordinal #1. */
         DosQueryProcAddr(hook_dll, 1UL, NULL, &hook_routine);
      }

      /* Set all the thread fields to nulls. */
      memset(&thread_area, 0x00, sizeof thread_area);
   
      rc = DosCreateEventSem(NULL, &thread_area.event_sem, 0UL, FALSE);
      if (rc == NO_ERROR)
      {
         file_thread = _beginthread(&FileThread, NULL, 48UL << 10, &thread_area);

         if (hook_dll != NULLHANDLE)
         {
            /* Set the input hook. */
            WinSetHook(anchor_block, HMQ_CURRENT, HK_INPUT, hook_routine, hook_dll);
         }
   
         /* Register a bocus window class. */
         WinRegisterClass(anchor_block, window_class,
               &MainWndProc, CS_SIZEREDRAW, sizeof(PVOID));
      
         /* Create an instance of that window class. */
         main_frame_window = WinCreateStdWindow(HWND_DESKTOP, WS_VISIBLE,
               &Frame_flags, window_class, window_title,
               WS_VISIBLE, NULLHANDLE, ID_MAIN_RESOURCE,
               &main_client_window);
      
         if (main_frame_window != NULLHANDLE)
         {
            /* Provide the message handlers with access to the thread control area. */
            WinSetWindowPtr(main_frame_window, 0UL, &thread_area);
            WinSetWindowPtr(main_client_window, 0UL, &thread_area);
         
            /* Your basic window dispatch loop. */
            while (WinGetMsg(anchor_block, &msg, NULLHANDLE, 0UL, 0UL))
               WinDispatchMsg(anchor_block, &msg);
         
            /* Clean up window resources. */
            WinDestroyWindow(main_frame_window);
         }
         else
            WinMessageBox(HWND_DESKTOP, HWND_DESKTOP,
               "Unable to create main window.", "Oh dear!", 0UL,
               MB_OK | MB_ERROR);
   
         /* Shut down the background thread, if it hasn't already died. */
         if (!thread_area.terminate_thread)
         {
            thread_area.terminate_thread = TRUE;
            DosPostEventSem(thread_area.event_sem);
         }

         if (hook_dll != NULLHANDLE)
         {
            /* Release the input hook. */
            WinReleaseHook(anchor_block, HMQ_CURRENT, HK_INPUT, hook_routine, hook_dll);
         }
   
         DosWaitThread(&file_thread, DCWW_WAIT);
   
         DosCloseEventSem(thread_area.event_sem);
      }
      else
      {
         sprintf(thread_area.window_text, "Unable to create event semaphore. rc = %lu", rc);
         WinMessageBox(HWND_DESKTOP, HWND_DESKTOP,
               thread_area.window_text, "Oops!", 0UL,
               MB_OK | MB_ERROR);
      }

      if (hook_dll != NULLHANDLE)
      {
         /* Release the DLL's memory. */
         DosFreeModule(hook_dll);
      }
   }
   else
   {
      sprintf(thread_area.window_text, "Unable to load hook DLL. rc = %lu\n%s", rc, crap);
      WinMessageBox(HWND_DESKTOP, HWND_DESKTOP,
            thread_area.window_text, "Drat!", 0UL,
            MB_OK | MB_ERROR);
   }

   /* Clean up PM resources. */
   WinDestroyMsgQueue(msg_queue);
   WinTerminate(anchor_block);

   return 0;
}
